/**
 * \file mkdistmap.cpp
 *
 * \author deong
 * \date 03/27/2008
 */

#include <iostream>
#include <iomanip>
#include <cmath>

using namespace std;

#define KEYS 68

/**
 * \brief main program producing the key-to-key distance mapping
 *
 * \author deong
 * \date 03/27/2008
 */
int main(int argc,char** argv)
{
    int row[KEYS];
    int col[KEYS];
    bool shifted[KEYS];
    int hand[KEYS];
    int dist[KEYS][KEYS];
    const char* keys = "1234567890-=\\`qwertyuiop[]asdfghjkl;'zxcvbnm,./!@#$%^&*()_+|~{}:\"<>?";

    /*
     * set the hand with which I type each key
     */
    for(int i=0; i<KEYS; i++)
    {
        if(i<=5 || (i>=14 && i<=18) || (i>=26 && i<=30) || (i>=37 && i<=41) || (i>=47 && i<=52))
        {
            hand[i]=0;
        }
        else
        {
            hand[i]=1;
        }
    }
    
    /*
     * figure out the row/column/shift status for each key
     */
    for(int i=0; i<KEYS; i++)
    {
        // top row of the keyboard
        if(i<=13)
        {
            row[i]=0;
            col[i]=i;
            shifted[i]=false;
        }

        // second row (qwerty plus brackets)
        else if(i>13 && i<=25)
        {
            row[i]=1;
            col[i]=i-14;        // only twelve keys; start at column 0
            shifted[i]=false;
        }

        // third row (home row)
        else if(i>25 && i<=36)
        {
            row[i]=2;
            col[i]=i-26;        // only eleven keys; start at column 0
            shifted[i]=false;
        }

        // fourth row
        else if(i>36 && i<=46)
        {
            row[i]=3;
            col[i]=i-37;        // only ten keys; start at column 0
            shifted[i]=false;
        }

        // shifted number row
        else if(i>46 && i<=60)
        {
            row[i]=0;
            col[i]=i-47;        // put into columns 0 through 13
            shifted[i]=true;
        }

        // curly braces
        else if(i>60 && i<=62)
        {
            row[i]=1;
            col[i]=i-51;        // put into columns 10 and 11
            shifted[i]=true;
        }

        // semicolon and double quote
        else if(i>62 && i<=64)
        {
            row[i]=2;
            col[i]=i-54;        // put into columns 9 and 10
            shifted[i]=true;
        }

        // <, >, and ?
        else
        {
            row[i]=3;
            col[i]=i-58;        // put into columns 7, 8, and 9
            shifted[i]=true;
        }
    }


    
    /*
     * now compute the distances between each pair of keys
     *
     * distance calculation is roughly as follows:
     *
     * start dist=0
     * if same hand
     *   rule 1
     *      add |row(x)-row(y)| + |col(x)-col(y)|
     *   rule 2
     *      if exactly one is shifted, add 5
     *      else if both are shifted, add 3
     * else
     *   rule 3
     *      if exactly one is shifted, add 8
     *      else if both are shifted, add 12
     *      
     * rule 4
     *   for k in {x,y}, add min(|col(k)-2|, |col(k)-3|, |col(k)-6|, |col(k)-7|)
     * rule 5
     *   if only one of col(x) and col(y) is > 11, add 2*(col(x)-10) or 2*(col(y)-10)
     *   if both are > 11, add min(col(x)-10,col(y)-10)
     * rule 6
     *   if row(x)==0, add 3
     *   if row(x)==1, add 1
     *   if row(x)==3, add 2
     */
    bool debug=false;
    for(int i=0; i<KEYS; i++)
    {
//         if(keys[i]=='d')
//             debug=true;
//         else
//             debug=false;
        
        for(int j=0; j<KEYS; j++)
        {
            int d=0;

            if(debug) cout << keys[i] << " -> " << keys[j] << endl;
            
            /* same hand */
            if(hand[i]==hand[j])
            {
                if(debug) cout << "same hand\n";
                
                /* rule 1 */
                d+=(static_cast<int>(fabs(row[i]-row[j])) +
                    static_cast<int>(fabs(col[i]-col[j])));

                if(debug) cout << "added manhattan distance...d=" << d << "\n";
                
                /* rule 2 */
                if(shifted[i]!=shifted[j])
                {
                    d+=5;
                    if(debug) cout << "one shifted...d=" << d << "\n";
                }
                else if(shifted[i]==true)
                {
                    d+=3;
                    if(debug) cout << "both shifted...d=" << d << "\n";
                }
            }
            
            /* alternate hands */
            else
            {
                if(debug) cout << "different hands\n";
                
                /* rule 3 */
                if(shifted[i]!=shifted[j])
                {
                    d+=8;
                    if(debug) cout << "one shifted...d=" << d << "\n";
                }
                else if(shifted[i]==true)
                {
                    d+=12;
                    if(debug) cout << "both shifted...d=" << d << "\n";
                }
            }

            /* penalized based on distance from home positions */

            /* rule 4 */
            if(hand[i]==0)
            {
                d+=static_cast<int>(min(fabs(col[i]-2),fabs(col[i]-3)));
                d+=static_cast<int>(min(fabs(col[j]-2),fabs(col[j]-3)));
                if(debug) cout << "left hand...adding distance to strong cols...d=" << d << "\n";
            }
            else
            {
                d+=static_cast<int>(min(fabs(col[i]-6),fabs(col[i]-7)));
                d+=static_cast<int>(min(fabs(col[j]-6),fabs(col[j]-7)));
                if(debug) cout << "right hand...adding distance to strong cols...d=" << d << "\n";
            }

            /* rule 5 */
            if(col[i]>11 && col[j]<=11)
            {
                d+=2*col[i]-10;
                if(debug) cout << "char 1 is way off to the right...d=" << d << "\n";
            }
            else if(col[j]>11 && col[i]<=11)
            {
                d+=2*col[j]-10;
                if(debug) cout << "char 2 is way off to the right...d=" << d << "\n";
            }
            else if(col[i]>11 && col[j]>11)
            {
                d+=min(col[i]-10,col[j]-10);
                if(debug) cout << "both characters are way off to the right...d=" << d << "\n";
            }

            /* rule 6 */
            if(row[i]==0 || row[j]==0)
            {
                d+=3;
                if(debug) cout << "number row...d=" << d << "\n";
            }
            else if(row[i]==1 || row[j]==1)
            {
                d+=1;
                if(debug) cout << "top alpha row...d=" << d << "\n";
            }
            else if(row[i]==3 || row[j]==3)
            {
                d+=2;
                if(debug) cout << "bottom row...d=" << d << "\n";
            }

            dist[i][j]=d;
            if(debug) cout << "total distance: " << d << "\n\n";
        }
    }

    /* print the distance matrix */
    cout << setw(4) << " ";
    for(int i=0; i<KEYS; i++)
    {
        cout << setw(4) << keys[i];
    }
    cout << "\n";
    
    for(int i=0; i<KEYS; i++)
    {
        cout << setw(4) << keys[i];
        for(int j=0; j<KEYS; j++)
        {
            cout << setw(4) << dist[i][j];
        }
        cout << "\n";
    }
    cout << endl;

    return 0;
}

    
